"""
    :author: Jinfen Li
    :url: https://github.com/JinfenLi
"""
import math
import torch
from torch import nn


class TokenAttention(nn.Module):
    def __init__(self, query_size, num_heads, hidden_size):
        super().__init__()
        self.num_attention_heads = num_heads
        self.attention_head_size = query_size
        self.all_head_size = self.num_attention_heads * self.attention_head_size

        self.query = nn.ModuleList([nn.Linear(self.attention_head_size, self.attention_head_size) for _ in range(self.num_attention_heads)])
        self.key = nn.Linear(hidden_size, query_size)
        self.value = nn.Linear(hidden_size, query_size)

        self.dense = nn.Linear(self.all_head_size, query_size)
        self.softmax = nn.Softmax(dim=-1)

    def transpose_for_scores(self, x):
        new_x_shape = x.size()[:-1] + (self.num_attention_heads, self.attention_head_size)
        x = x.view(*new_x_shape)
        return x.permute(0, 2, 1, 3)

    def forward(self, hidden_states, query_states):
        """

        Args:
            hidden_states: the word vec of the covered terminal words: [Batch_size, Seq_length, Hidden_size]
            query_states: the phrase vec of the covered terminal words: [Batch_size, Num_of_heads x Seq_length, Query_size]

        Returns:

        """
        new_query_shape = query_states.size()[:-2] + (hidden_states.size()[1], self.num_attention_heads, self.attention_head_size)
        query_states = query_states.view(*new_query_shape) # [Batch_size x Seq_length x Num_of_heads x Head_size]
        mixed_query_layer = torch.stack([self.query[i](query_states[:, :, i]) for i in range(self.num_attention_heads)])  # [Num_of_heads x Batch_size x Seq_length x Head_size]
        mixed_key_layer = self.key(hidden_states)  # [Batch_size x Seq_length x Head_size]
        mixed_value_layer = self.value(hidden_states)  # [Batch_size x Seq_length x Head_size]

        attention_scores = torch.stack([torch.matmul(mixed_query_layer[i], mixed_key_layer.transpose(-1,
                                                                            -2)) for i in range(self.num_attention_heads)])  # [Num_of_heads x Batch_size x Seq_length x Seq_length]

        attention_scores = attention_scores / math.sqrt(
            self.attention_head_size)  # [Num_of_heads x Batch_size x Seq_length x Seq_length]
        attention_probs = nn.Softmax(dim=-1)(attention_scores)  # [Num_of_heads x Batch_size x Seq_length x Seq_length]
        context_layer = torch.matmul(attention_probs,
                                     mixed_value_layer)  # [Num_of_heads x Batch_size x Seq_length x Head_size]

        context_layer = context_layer.permute(1, 2, 0, 3).contiguous()  # [Batch_size x Seq_length x Num_of_heads x Head_size]
        new_context_layer_shape = context_layer.size()[:-2] + (
            self.all_head_size,)  # [Batch_size x Seq_length x Hidden_size]
        context_layer = context_layer.view(*new_context_layer_shape)  # [Batch_size x Seq_length x Hidden_size]

        output = self.dense(context_layer)
        # pick the one with the highest attention score
        output = torch.max(output, dim=1)[0]

        return output


"""
    :author: Jinfen Li
    :url: https://github.com/JinfenLi
"""
import socket
import subprocess
from collections import defaultdict as ddict
import requests
from tqdm import tqdm

def remap_token_id(reference_offset, parsed_offset, ref_offset_tid = 1):
    """

    Args:
        reference_offset: offset generated by original model
        parsed_offset: offset generated by stanfordnlp
        ref_offset_tid: initial reference_offset token id, the first reference_offset is for [CLS]

    Returns: {new_token_dict: [start_ref_tid, end_ref_tid)}, end_ref_tid is not inclusive

    """
    parsed_offset_tid = 0
    new_token_dict = {}
    start = ref_offset_tid
    while parsed_offset_tid < len(parsed_offset):
        if reference_offset[ref_offset_tid][0] == 0 and reference_offset[ref_offset_tid][1] == 0:
            break
        elif parsed_offset[parsed_offset_tid][1] >= reference_offset[ref_offset_tid][1]:
            # find the first reference_offset that exceed parsed_offset
            while ref_offset_tid < (len(reference_offset) - 1) and parsed_offset[parsed_offset_tid][1] > reference_offset[ref_offset_tid][
                1]:
                ref_offset_tid += 1
            if parsed_offset_tid > 0:
                start = new_token_dict[parsed_offset_tid - 1][1]
            end = ref_offset_tid + 1
            new_token_dict[parsed_offset_tid] = [start, end]
            parsed_offset_tid += 1
        elif parsed_offset[parsed_offset_tid][1] < reference_offset[ref_offset_tid][1]:
            if parsed_offset_tid > 0:
                start = new_token_dict[parsed_offset_tid - 1][1]
            end = ref_offset_tid + 1
            new_token_dict[parsed_offset_tid] = [start, end]
            parsed_offset_tid += 1

    return new_token_dict, ref_offset_tid + 1

def parse_senti_tree(senti_tree_str, tid=0):
    """
    Args:
        senti_tree_str: a string of sentiment tree
        tid: initial token id
    Returns: phrase_dict: {'sentiment': [1, 2], 'tokens': [[1, 2], [1, 2]]}

    """
    senti_nodes = [chars.strip() for chars in senti_tree_str.split('\r\n')]
    phrase_dict = ddict(list)
    # stack: [{label: str, sentiment: int, span: list, left_count: int, right_count: int}]

    stack = []

    for node in senti_nodes:
        labels = [l.split("|")[0] for l in node.split('(') if l]
        sentiments = [int(s) for s in re.findall(r"sentiment=(\d)", node)]

        if len(labels) == 3:
            stack.append({"label": labels[0], "sentiment": sentiments[0], "span": [tid, tid + 1], "left_count": node.count("("), "right_count": node.count(")")})
            tid += 2
        elif len(labels) == 2:
            stack.append(
                {"label": labels[0], "sentiment": sentiments[0], "span": [tid], "left_count": node.count("("),
                 "right_count": node.count(")")})
            tid += 1
        elif len(labels) == 1:
            stack.append({"label": labels[0], "sentiment": sentiments[0], "span": [tid] if node.count(")") > 0 else None, "left_count": node.count("("), "right_count": node.count(")")})
            if node.count(")") > 0:
                tid += 1
        # if right_count > left_count, means rolling back
        while stack and stack[-1]["right_count"] > stack[-1]["left_count"]:
            element = stack.pop()
            stack[-1]["right_count"] = stack[-1]["right_count"] + element["right_count"]
            stack[-1]["left_count"] = stack[-1]["left_count"] + element["left_count"]
            # the last element is the left node
            if stack[-1]["span"] is not None and len(stack[-1]["span"]) == 2:
                continue
            stack[-1]["span"] = [stack[-1]["span"][0] if stack[-1]["span"] is not None else element["span"][0], tid - 1]
            if stack[-1]["label"] in ['@S', 'S', 'ROOT']:
                phrase_dict['sentiment'].append(stack[-1]["sentiment"])
                phrase_dict['tokens'].append(stack[-1]["span"])

    return phrase_dict

def sentiment_tree(texts, num_examples, original_offsets, resource_dir):
    sents_with_tokens_list = []
    available_port = get_CoreNLPClient(resource_dir)
    # with CoreNLPClient(be_quiet=False, annotators=['tokenize', 'sentiment'], properties={'tokenize.codepoint': 'true'}, timeout=60000,
    #                    memory='16G', output_format='json') as client:

    for idx in tqdm(range(num_examples), desc=f"Building dataset"):
        # new_senti_dict = ddict(list)
        text = texts[idx]
        original_offset = original_offsets[idx]
        # ann = client.annotate(text.strip())
        ann = annotate_text(text.strip(), available_port)
        cur_start_tid = 0
        result_dict = ddict(list)
        new_token_dict, phrase_dict = {}, {}
        cur_start_ref_tid = 1
        for sent_dict in ann['sentences']:

            sent_dict['offset'] = [(x['characterOffsetBegin'], x['characterOffsetEnd']) for x in sent_dict['tokens']]
            # original_token_idx = [x['index'] for x in sent_dict['tokens']]
            remap_token_dict, cur_start_ref_tid = remap_token_id(original_offset, sent_dict['offset'], cur_start_ref_tid)
            new_token_dict.update({k + cur_start_tid: v for k, v in remap_token_dict.items()})
            phrase_dict = parse_senti_tree(sent_dict['sentimentTree'], cur_start_tid)
            phrase_dict['tokens'] = [[new_token_dict[x[0]][0], new_token_dict[x[1]][1]] for x in phrase_dict['tokens']]
            result_dict['spans'].extend(phrase_dict['tokens'])
            result_dict['sentiment'].extend(phrase_dict['sentiment'])
            cur_start_tid = max(new_token_dict.keys()) + 1
        sents_with_tokens_list.append(result_dict)


    return sents_with_tokens_list


def get_CoreNLPClient(resource_dir):
    def find_available_port(start_port, max_attempts=10):
        for port in range(start_port, start_port + max_attempts):
            try:
                with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as s:
                    s.bind(("127.0.0.1", port))
                return port
            except OSError:
                pass
        return None

    available_port = find_available_port(start_port=9000)

    if available_port:
        print("Available port for CoreNLPClient:", available_port)
    else:
        print("No available port  for CoreNLPClient found")
    # Define the command to start the CoreNLP server
    corenlp_command = [
        "java", "-mx16g", "-cp", f"{resource_dir}/stanford-corenlp-4.5.4/*",
        "edu.stanford.nlp.pipeline.StanfordCoreNLPServer",
        "-port", f"{available_port}", "-timeout", "60000", "-threads", "4", "-maxCharLength", "100000", "-quiet",
        "True",
        "-annotators", "tokenize,sentiment",
        "properties", "{'tokenize.codepoint': 'true'}"
    ]

    # Start the CoreNLP server
    subprocess.Popen(corenlp_command)
    return available_port

def annotate_text(text, available_port):

    corenlp_url = f"http://localhost:{available_port}"
    properties = {
        "annotators": "parse,sentiment",
        "properties": "{'tokenize.codepoint': 'true'}",
        "outputFormat": "json"
    }

    response = requests.post(f"{corenlp_url}/?properties={properties}", data=text.encode("utf-8"))
    json_response = response.json()
    return json_response



from lightning.pytorch import Callback
class BestPerformance(Callback):

    def __init__(self, monitor, mode):
        super().__init__()

        self.monitor = monitor
        assert monitor.split('_')[0] == 'dev'
        self.test_monitor = '_'.join(['test'] + monitor.split('_')[1:])

        self.mode = mode
        assert mode in ['max', 'min']



import torch
from torch import nn

from transformers import AutoModel, AutoTokenizer
from multi_emotion.model.base_model import BaseModel
from multi_emotion.model.attention import TokenAttention
from multi_emotion.utils.data import English_Hashtag_Voc_Size, Spanish_Hashtag_Voc_Size
from multi_emotion.utils.losses import calc_task_loss
from multi_emotion.utils.metrics import init_best_metrics, init_perf_metrics
from multi_emotion.utils.optim import setup_optimizer_params, setup_scheduler, freeze_layers
from multi_emotion.utils.logging import log_step_losses, log_epoch_losses, log_epoch_metrics
from multi_emotion.utils.utils import generate_output_file


class MultiEmoModel(BaseModel):
    def __init__(self, arch: str, use_hashtag: bool = True, use_senti_tree: bool = True, use_emo_cor: bool = True,
                 hashtag_emb_dim: int = 0,
                 phrase_emb_dim: int = 0, senti_emb_dim: int = 0, max_length: int = 0, num_classes: int = None,
                 dataset: str = None, num_freeze_layers: int = 0, freeze_epochs=-1, neg_weight=1,
                 save_outputs: str = None, exp_id: str = None,
                 measure_attrs_runtime: bool = False, optimizer: torch.optim.Optimizer = None,
                 scheduler: torch.optim.lr_scheduler = None, **kwargs):


        super().__init__()
        # self.optimizer = optimizer
        # self.scheduler = scheduler
        self.save_hyperparameters(logger=False)

        self.arch = arch
        self.dataset = dataset
        self.num_classes = num_classes
        self.max_length = max_length

        self.freeze_epochs = freeze_epochs
        self.neg_weight = neg_weight
        self.use_hashtag = use_hashtag
        self.use_senti_tree = use_senti_tree
        self.use_emo_cor = use_emo_cor
        self.hashtag_emb_dim = hashtag_emb_dim
        self.phrase_emb_dim = phrase_emb_dim
        self.senti_emb_dim = senti_emb_dim

        self.best_metrics = init_best_metrics()
        self.perf_metrics = init_perf_metrics(num_classes)

        self.register_buffer('empty_tensor', torch.LongTensor([]))
        # if num_classes == 2:
        #     self.register_buffer('class_weights', torch.FloatTensor([neg_weight, 1]))
        # else:
        #     self.class_weights = None

        self.tokenizer = AutoTokenizer.from_pretrained(arch)
        self.task_encoder = AutoModel.from_pretrained(arch)
        task_head_input_size = self.task_encoder.config.hidden_size

        hashtag_voc_size = English_Hashtag_Voc_Size if dataset not in ['se_spanish', 'se_arabic'] else (
            Spanish_Hashtag_Voc_Size if dataset == 'se_spanish' else None)
        self.hashtag_encoder = nn.Embedding(hashtag_voc_size, self.hashtag_emb_dim) if (
                    use_hashtag and hashtag_voc_size is not None) else None
        if use_hashtag:
            task_head_input_size += self.hashtag_emb_dim

        # sentiment for each phrase node
        self.senti_encoder = nn.Embedding(5, self.senti_emb_dim) if (
            use_senti_tree) else None
        self.phrase_encoder = nn.Linear(self.task_encoder.config.hidden_size,
                                        self.phrase_emb_dim) if use_senti_tree else None
        self.phrase_head = nn.Linear(self.phrase_emb_dim, self.phrase_emb_dim) if use_senti_tree else None
        self.tok_attn = TokenAttention(self.phrase_emb_dim, 1,
                                       self.task_encoder.config.hidden_size) if use_senti_tree else None
        if use_senti_tree:
            task_head_input_size += self.phrase_emb_dim + self.senti_emb_dim

        self.task_head = nn.Linear(
            task_head_input_size,
            num_classes
        )
        self.cor_head = None
        if use_emo_cor:
            self.task_head = nn.Linear(
                task_head_input_size,
                task_head_input_size
            )
            self.cor_head = nn.Bilinear(task_head_input_size, task_head_input_size, num_classes)

        # self.sigmoid = nn.Sigmoid()

        assert num_freeze_layers >= 0
        if num_freeze_layers > 0:
            freeze_layers(self, num_freeze_layers)

        self.model_dict = {
            'task_encoder': self.task_encoder,
            'task_head': self.task_head,
            'hashtag_encoder': self.hashtag_encoder,
            'phrase_encoder': self.phrase_encoder,
            'phrase_head': self.phrase_head,
            'tok_attn': self.tok_attn,
            'senti_encoder': self.senti_encoder,
            'cor_head': self.cor_head,
            # 'sigmoid': self.sigmoid

        }

        # if save_outputs:
        #     assert exp_id is not None
        self.save_outputs = save_outputs
        self.exp_id = exp_id

        self.measure_attrs_runtime = measure_attrs_runtime

    def generate_phrase_embs(self, hidden_states, phrase_ids, sentiment_ids):
        senti_enc = self.senti_encoder(sentiment_ids)
        enc = self.phrase_encoder(hidden_states)
        enc = self.phrase_head(enc)
        total_encs = []

        for i, cur_pids in enumerate(phrase_ids):
            total_enc = []
            for j, pid in enumerate(cur_pids):
                if pid[1] == 0:
                    phrase_encs = enc[i, 0].unsqueeze(0).unsqueeze(0)
                    # get neutral sentiment
                    senti_encs = self.senti_encoder.weight[2].unsqueeze(0)
                    word_encs = hidden_states[i, 0].unsqueeze(0).unsqueeze(0)

                else:
                    phrase_encs = enc[i, pid[0]: pid[1]].unsqueeze(0)
                    senti_encs = senti_enc[i, j].unsqueeze(0)
                    word_encs = hidden_states[i, pid[0]: pid[1]].unsqueeze(0)

                cur_phrase_encs = self.tok_attn(word_encs, phrase_encs)
                total_enc.append(torch.cat([cur_phrase_encs, senti_encs], dim=1))
            cur_total_enc = torch.stack(total_enc).mean(dim=0)
            total_encs.append(cur_total_enc)
        # take the average of total_enc
        total_encs = torch.stack(total_encs).squeeze(1)
        return total_encs

    def forward(self, inputs, attention_mask, hashtag_inputs, phrase_inputs, sentiment_ids):

        enc = self.task_encoder(input_ids=inputs, attention_mask=attention_mask)
        task_head_inputs = enc.pooler_output
        if hashtag_inputs is not None:
            hashtag_enc = self.hashtag_encoder(hashtag_inputs)
            hashtag_enc = torch.mean(hashtag_enc, dim=1)
            task_head_inputs = torch.cat([task_head_inputs, hashtag_enc], dim=1)
        if phrase_inputs is not None:
            phrase_enc = self.generate_phrase_embs(enc.last_hidden_state, phrase_inputs, sentiment_ids)
            task_head_inputs = torch.cat([task_head_inputs, phrase_enc], dim=1)

        logits = self.task_head(task_head_inputs)
        if self.use_emo_cor:
            logits = self.cor_head(logits, logits)

        return logits

    def run_step(self, batch, split, batch_idx):

        input_ids = batch['input_ids']
        attn_mask = batch['attention_mask']
        hashtag_ids = batch['hashtag_ids'] if self.use_hashtag else None
        phrase_ids = batch['phrase_span_ids'] if self.use_senti_tree else None
        sentiment_ids = batch['sentiment_ids']
        targets = batch['label']
        eval_split: str = batch['split']
        if split == 'train':
            assert split == eval_split

        ret_dict, loss_dict, metric_dict = {}, {}, {}

        logits = self.forward(input_ids, attn_mask, hashtag_ids, phrase_ids, sentiment_ids)
        probs = logits.sigmoid()
        if targets is not None:
            loss = calc_task_loss(logits, targets) if targets is not None else None
            loss_dict['loss'] = loss
            # Log step losses
            ret_dict = log_step_losses(self, loss_dict, ret_dict, eval_split)

        ret_dict['targets'] = targets.detach() if targets is not None else None
        ret_dict['probs'] = probs.detach()
        ret_dict['eval_split'] = eval_split
        ret_dict['input_ids'] = input_ids.detach()
        return ret_dict

    def aggregate_epoch(self, outputs, split):
        if outputs['targets'] is not None:
            log_epoch_losses(self, outputs, outputs['eval_split'][0])  # Log epoch losses
            log_epoch_metrics(self, outputs, outputs['eval_split'][0])  # Log epoch metrics
        results = None
        if outputs['eval_split'][0] == 'pred':
            input_ids = outputs['input_ids']
            targets = outputs['targets']
            probs = outputs['probs']
            # out_data = calc_preds(logits)
            probabilities = probs.cpu().detach()

            results = generate_output_file(self.dataset, self.tokenizer, input_ids, probabilities, self.save_outputs, targets)
        return results



    def configure_optimizers(self):

        optimizer_params = setup_optimizer_params(self.model_dict, self.hparams.optimizer, self.use_hashtag,
                                                  self.use_senti_tree, self.use_emo_cor)
        self.hparams.optimizer.keywords['lr'] = self.hparams.optimizer.keywords['lr'] * self.trainer.world_size
        optimizer = self.hparams.optimizer(params=optimizer_params)
        if self.hparams.scheduler['lr_scheduler'] == 'linear_with_warmup':
            scheduler = setup_scheduler(self.hparams.scheduler, self.total_steps, optimizer)
            return [optimizer], [scheduler]
        elif self.hparams.scheduler['lr_scheduler'] == 'fixed':
            return [optimizer]
        else:
            raise NotImplementedError



English_Hashtag_Voc_Size = 32389 + 1
Spanish_Hashtag_Voc_Size = 2037 + 1

dataset_info = {
    'se_english': {
        'train': ['train', 6838],
        'dev': ['dev', 886],
        'test': ['test', 3259],
        'num_classes': 11,
        'classes': ["anger", "anticipation", "disgust", "fear", "joy", "love", "optimism", "pessimism", "sadness",
                    "surprise", "trust"],
        'max_length': {
            'bert-base-uncased': 128,
        },
        'num_special_tokens': 2,
    }
}

benchmark_datasets = ['se_english', 'se_arabic', 'se_spanish']

monitor_dict = {
    'se_english': 'dev_macro_f1_metric_epoch',
    'se_arabic': 'dev_macro_f1_metric_epoch',
    'se_spanish': 'dev_macro_f1_metric_epoch',
}

data_keys = ['item_idx', 'input_ids', 'hashtag_ids', 'attention_mask', 'tree', 'label']




import getpass, socket
from typing import Any, List
import torch
from lightning_utilities.core.rank_zero import rank_zero_only
from lightning.pytorch.loggers import CSVLogger
from multi_emotion.utils.metrics import get_step_metrics, get_epoch_metrics
import logging



log = logging.getLogger(__name__)

def get_username():
    return getpass.getuser()

def flatten_cfg(cfg: Any) -> dict:
    if isinstance(cfg, dict):
        ret = {}
        for k, v in cfg.items():
            flatten: dict = flatten_cfg(v)
            ret.update({
                f"{k}/{f}" if f else k: fv
                for f, fv in flatten.items()
            })
        return ret
    return {"": cfg}

def get_logger(name=__name__, level=logging.INFO) -> logging.Logger:
    logger = logging.getLogger(name)
    logger.setLevel(level)
    return logger

def get_csv_logger(
    cfg, logger, save_dir, name, offline
):
    csv_logger = CSVLogger(
        save_dir=save_dir,
        name=name
    )

    return csv_logger



def log_data_to_model(model_class, data, data_name, data_type, suffix, split, ret_dict=None, detach_data=True):

    data_key = 'loss' if f'{data_name}_{data_type}' == 'total_loss' else f'{data_name}_{data_type}'
    model_class.log(f'{split}_{data_key}_{suffix}', data.detach(), prog_bar=True, sync_dist=(split != 'train'))
    if ret_dict is not None:
        ret_dict[data_key] = data.detach() if detach_data else data
    
    return ret_dict

def log_step_losses(model_class, loss_dict, ret_dict, split):
    ret_dict = log_data_to_model(model_class, loss_dict['loss'], 'total', 'loss', 'step', split, ret_dict, detach_data=False)
    return ret_dict


def log_epoch_losses(model_class, outputs, split):
    loss = outputs['loss'].mean()
    log_data_to_model(model_class, loss, 'total', 'loss', 'epoch', split, ret_dict=None)

def log_epoch_metrics(model_class, outputs, split):
    probs = outputs['probs']
    targets = outputs['targets']
    get_step_metrics(probs, targets, model_class.perf_metrics)
    perf_metrics = get_epoch_metrics(model_class.perf_metrics)

    log_data_to_model(model_class, perf_metrics['acc'], 'acc', 'metric', 'epoch', split, ret_dict=None)
    log_data_to_model(model_class, perf_metrics['macro_f1'], 'macro_f1', 'metric', 'epoch', split, ret_dict=None)

    if model_class.num_classes == 2:
        log_data_to_model(model_class, perf_metrics['binary_f1'], 'binary_f1', 'metric', 'epoch', split, ret_dict=None)


    if 'delta' in outputs.keys():
        delta = torch.abs(outputs['delta']).mean()
        log_data_to_model(model_class, delta, 'convergence_delta', 'metric', 'epoch', split, ret_dict=None)

@rank_zero_only
def log_hyperparameters(object_dict: dict) -> None:
    """Controls which config parts are saved by lightning loggers.

    Additionally saves:
    - Number of model parameters
    """

    hparams = {}

    cfg = object_dict["cfg"]
    model = object_dict["model"]
    trainer = object_dict["trainer"]

    if not trainer.logger:
        log.warning("Logger not found! Skipping hyperparameter logging...")
        return

    hparams["model"] = cfg["model"]

    # save number of model parameters
    hparams["model/params/total"] = sum(p.numel() for p in model.parameters())
    hparams["model/params/trainable"] = sum(
        p.numel() for p in model.parameters() if p.requires_grad
    )
    hparams["model/params/non_trainable"] = sum(
        p.numel() for p in model.parameters() if not p.requires_grad
    )

    hparams["data"] = cfg["data"]
    hparams["trainer"] = cfg["trainer"]

    hparams["callbacks"] = cfg.get("callbacks")
    hparams["extras"] = cfg.get("extras")

    hparams["task_name"] = cfg.get("task_name")
    hparams["tags"] = cfg.get("tags")
    hparams["ckpt_path"] = cfg.get("ckpt_path")
    hparams["seed"] = cfg.get("seed")

    # send hparams to all loggers
    for logger in trainer.loggers:
        logger.log_hyperparams(hparams)


"""
handle changes to the hydra config
"""
from pathlib import Path

import rich.syntax
import rich.tree
from omegaconf import DictConfig, ListConfig, OmegaConf
from lightning.pytorch.utilities import rank_zero_only

def fail_on_missing(cfg: DictConfig) -> None:
    if isinstance(cfg, ListConfig):
        for x in cfg:
            fail_on_missing(x)
    elif isinstance(cfg, DictConfig):
        for _, v in cfg.items():
            fail_on_missing(v)


def pretty_print(
        cfg: DictConfig,
        fields=(
            "dataset",
            "model",
            "logger",
            "trainer",
            "setup",
            "training",
        )
):
    style = "dim"
    tree = rich.tree.Tree(":gear: CONFIG", style=style, guide_style=style)

    for field in fields:
        branch = tree.add(field, style=style, guide_style=style)

        config_section = cfg.get(field)
        branch_content = str(config_section)
        if isinstance(config_section, DictConfig):
            branch_content = OmegaConf.to_yaml(config_section, resolve=True)

        branch.add(rich.syntax.Syntax(branch_content, "yaml"))

    # others defined in root
    others = tree.add("others", style=style, guide_style=style)
    for var, val in OmegaConf.to_container(cfg, resolve=True).items():
        if not var.startswith("_") and var not in fields:
            others.add(f"{var}: {val}")

    rich.print(tree)


@rank_zero_only
def touch(cfg: DictConfig) -> None:
    # sanity check
    assert Path(cfg.data.data_path).exists(), f"datapath {cfg.data.data_path} not exist"

    if cfg.training.finetune_ckpt:
        assert cfg.training.ckpt_path
    if cfg.training.evaluate_ckpt:
        assert cfg.training.ckpt_path
        assert cfg.training.eval_splits != "all"

    if cfg.debug:
        # for DEBUG purposes only
        cfg.trainer.limit_train_batches = 1
        cfg.trainer.limit_val_batches = 1
        cfg.trainer.limit_test_batches = 1
        cfg.trainer.max_epochs = 1
        # for DEBUG purposes only

    fail_on_missing(cfg)
    pretty_print(cfg)



"""
    :author: Jinfen Li
    :url: https://github.com/JinfenLi
"""
import io
import logging
import os
import tarfile
import torch
import yaml
from lightning.pytorch import Trainer
from tqdm import tqdm
from transformers import AutoTokenizer
from collections import defaultdict as ddict

from .model.lm import MultiEmoModel
from .data.data import DataModule
from .utils.sentiTree import data_utils
from .utils.utils import nrc_hashtag_lexicon, preprocess_dataset, transform_data, update_dataset_dict

logging.basicConfig(level=logging.INFO)
logger = logging.getLogger(__name__)

def get_configuration():
    path_to_exist = os.path.join(os.path.dirname(__file__), f"model_dependencies")
    os.makedirs(path_to_exist, exist_ok=True)
    if not os.path.exists(os.path.join(path_to_exist, "multi-emo-bert.tar.gz")):
        logger.info(f"Downloading model configuration files to {path_to_exist} ...")
        torch.hub.download_url_to_file("https://pytorch-libs.s3.us-east-2.amazonaws.com/multi-emo-bert.tar.gz",
                                       os.path.join(path_to_exist, "multi-emo-bert.tar.gz"))

        with tarfile.open(os.path.join(path_to_exist, "multi-emo-bert.tar.gz"), 'r:gz') as tar:
            tar.extractall(path=path_to_exist)
    with open(os.path.join(path_to_exist, "multi-emo-bert", 'config.yaml'), 'r') as f:
        config = yaml.safe_load(f)
    with open(os.path.join(path_to_exist, "multi-emo-bert", 'hashtag_lexicon.txt'), 'r') as f:
        hashtag_lexicon = f.readlines()

    ckpt_path = os.path.join(path_to_exist, "multi-emo-bert", 'pytorch_model.ckpt')
    return config, hashtag_lexicon, ckpt_path


def convert_text_to_features(texts, tokenizer, max_length, hashtag_dict, use_senti_tree=False):
    """
    convert text to features
    Args:
        texts: text to convert
        tokenizer: tokenizer
        max_length: max length of text
        use_senti_tree: whether to use senti tree
        labels: labels can be None or List[List[]]
    """
    dataset_dict = ddict(list)
    for idx in tqdm(range(0, len(texts)), desc=f'Building dataset'):

        label = None
        text = preprocess_dataset(texts[idx])
        input_ids, hashtag_inputs, offset_mapping = transform_data(tokenizer, hashtag_dict, text, max_length)
        dataset_dict = update_dataset_dict(idx, dataset_dict, input_ids, hashtag_inputs, max_length, tokenizer, text, offset_mapping,
                                           label)
    if use_senti_tree:
        dataset_dict['tree'] = data_utils.sentiment_tree(dataset_dict['truncated_texts'],
                                              len(texts),
                                              dataset_dict['offsets'],
                                              max_length)
    return dataset_dict

def load_checkpoint(model, ckpt_path):

    buffer = io.BytesIO()
    torch.save(ckpt_path, buffer)
    buffer.seek(0)  # Reset the buffer position to the beginning
    checkpoint = torch.load(buffer)
    model = model.load_from_checkpoint(checkpoint, strict=False, map_location=torch.device('cpu') if not torch.cuda.is_available() else None)
    logger.info(f"Loaded checkpoint for evaluation from {ckpt_path}")
    return model

def predict(texts: list):
    """

    Args:
        texts: a list of texts

    Returns:
        [{text: [text1, text2, ...], label: [emotion1, emotion2, ...], probability: [{anger: 0.1}, {anticipation: 0.2}, {...}]}]

    """
    assert isinstance(texts, list), "texts must be a list of texts"
    config, hashtag_lexicon, ckpt_path = get_configuration()
    max_length = config['max_length']
    arch = config['arch']
    use_hashtag = config['use_hashtag']
    use_senti_tree = config['use_senti_tree']
    phrase_num = config['phrase_num']
    use_emo_cor = config['use_emo_cor']
    hashtag_emb_dim = config['hashtag_emb_dim']
    phrase_emb_dim = config['phrase_emb_dim']
    senti_emb_dim = config['senti_emb_dim']
    num_classes = config['num_classes']

    tokenizer = AutoTokenizer.from_pretrained(arch, strip_accents=False)
    hashtag_vocab = nrc_hashtag_lexicon(hashtag_lexicon)
    hashtag_dict = {h: i + 1 for i, h in enumerate(hashtag_vocab)}
    dataset_dict = convert_text_to_features(texts, tokenizer, max_length, hashtag_dict, use_senti_tree=use_senti_tree)
    dm = DataModule(use_hashtag, use_senti_tree, phrase_num)
    dm.setup(dataset=dataset_dict)
    loader = dm.predict_dataloader()

    model = MultiEmoModel(arch, use_hashtag, use_senti_tree, use_emo_cor,
                 hashtag_emb_dim,
                 phrase_emb_dim, senti_emb_dim, max_length, num_classes)
    model = load_checkpoint(model, ckpt_path)

    trainer = Trainer(logger=False)
    trainer.predict(model=model, dataloaders=loader)
    test_results = trainer.lightning_module.results
    return test_results




from typing import Optional
from lightning import LightningModule
import torch

class BaseModel(LightningModule):
    def __init__(self):
        super().__init__()
        # update in `setup`
        self.total_steps = None
        self.training_step_outputs = []
        self.validation_step_outputs = [[], []]
        self.test_step_outputs = []
        self.predict_step_outputs = []
        self.results = None


    def forward(self, **kwargs):
        raise NotImplementedError

    def calc_loss(self, preds, targets):
        raise NotImplementedError

    def calc_acc(self, preds, targets):
        raise NotImplementedError

    def run_step(self, batch, split, batch_idx):
        raise NotImplementedError

    def aggregate_epoch(self, outputs, split):
        raise NotImplementedError

    def collater(self, outputs):
        total_outputs = {
            'loss': torch.stack([x['loss'] for x in outputs]) if 'loss' in outputs[0] else None,
            'probs': torch.cat([x['probs'] for x in outputs]),
            'targets': torch.cat([x['targets'] for x in outputs]) if outputs[0]['targets'] is not None else None,
            'eval_split': [x['eval_split'] for x in outputs],
            'input_ids': torch.cat([x['input_ids'] for x in outputs]),

        }

        return total_outputs

    def training_step(self, batch, batch_idx):
        # # freeze encoder for initial few epochs based on p.freeze_epochs
        # if self.current_epoch < self.freeze_epochs:
        # 	freeze_net(self.text_encoder)
        # else:
        # 	unfreeze_net(self.text_encoder)
        # print("training step")
        ret_dict = self.run_step(batch, 'train', batch_idx)
        # loss = ret_dict['loss']
        self.training_step_outputs.append(ret_dict)
        return ret_dict


    def on_train_epoch_end(self):
        training_step_outputs = self.collater(self.training_step_outputs)
        self.aggregate_epoch(training_step_outputs, 'train')
        self.training_step_outputs.clear()

    def validation_step(self, batch, batch_idx, dataloader_idx=0):
        # print("validation_step step")
        assert dataloader_idx in [0, 1]
        eval_splits = {0: 'dev', 1: 'test'}
        ret_dict = self.run_step(batch, eval_splits[dataloader_idx], batch_idx)
        self.validation_step_outputs[dataloader_idx].append(ret_dict)
        return ret_dict

    def on_validation_epoch_end(self):
        for dl_idx in range(len(self.validation_step_outputs)):
            validation_step_outputs = self.collater(self.validation_step_outputs[dl_idx])
            self.aggregate_epoch(validation_step_outputs, 'dev')
            self.validation_step_outputs[dl_idx].clear()

        monitor = self.trainer.callbacks[0].monitor
        test_monitor = self.trainer.callbacks[0].test_monitor
        if self.trainer.callbacks[0].mode == 'max':
            if self.best_metrics['dev_best_perf'] == None:
                assert self.best_metrics['test_best_perf'] == None
                self.best_metrics['dev_best_perf'] = -float('inf')

            if self.trainer.callback_metrics[monitor] > self.best_metrics['dev_best_perf']:
                self.best_metrics['dev_best_perf'] = self.trainer.callback_metrics[monitor]
                self.best_metrics['test_best_perf'] = self.trainer.callback_metrics[test_monitor]
                self.best_metrics['best_epoch'] = self.trainer.current_epoch
        else:
            if self.best_metrics['dev_best_perf'] == None:
                assert self.best_metrics['test_best_perf'] == None
                self.best_metrics['dev_best_perf'] = float('inf')

            if self.trainer.callback_metrics[monitor] < self.best_metrics['dev_best_perf']:
                self.best_metrics['dev_best_perf'] = self.trainer.callback_metrics[monitor]
                self.best_metrics['test_best_perf'] = self.trainer.callback_metrics[test_monitor]
                self.best_metrics['best_epoch'] = self.trainer.current_epoch

        self.log('dev_best_perf', self.best_metrics['dev_best_perf'], prog_bar=True, sync_dist=True)
        self.log('test_best_perf', self.best_metrics['test_best_perf'], prog_bar=True, sync_dist=True)
        self.log('best_epoch', self.best_metrics['best_epoch'], prog_bar=True, sync_dist=True)

    def test_step(self, batch, batch_idx):
        self.test_step_outputs.append(self.run_step(batch, 'test', batch_idx))
        return self.run_step(batch, 'test', batch_idx)

    def on_test_epoch_end(self):
        test_step_outputs = self.collater(self.test_step_outputs)
        self.aggregate_epoch(test_step_outputs, 'test')
        self.test_step_outputs.clear()
        # self.results = results
        # return results

    def predict_step(self, batch, batch_idx, dataloader_idx=0):
        self.predict_step_outputs.append(self.run_step(batch, 'pred', batch_idx))
        return self.run_step(batch, 'pred', batch_idx)

    def on_predict_epoch_end(self):
        predict_step_outputs = self.collater(self.predict_step_outputs)
        results = self.aggregate_epoch(predict_step_outputs, 'pred')
        self.predict_step_outputs.clear()
        self.results = results
        return results

    def setup(self, stage: Optional[str] = None):
        """calculate total steps"""
        if stage == 'fit':
            # Get train dataloader
            train_loader = self.trainer.datamodule.train_dataloader()
            ngpus = self.trainer.num_devices

            # Calculate total steps
            eff_train_batch_size = (self.trainer.datamodule.train_batch_size *
                                    max(1, ngpus) * self.trainer.accumulate_grad_batches)
            # assert eff_train_batch_size == self.trainer.datamodule.eff_train_batch_size
            self.total_steps = int(
                (len(train_loader.dataset) // eff_train_batch_size) * float(self.trainer.max_epochs))

    def configure_optimizers(self):
        raise NotImplementedError


import os
from pathlib import Path
import pickle
from lightning.pytorch import LightningDataModule
import torch
from torch.utils.data import DataLoader, Dataset
from tqdm import tqdm
from multi_emotion.utils.data import data_keys


class DataModule(LightningDataModule):

    def __init__(self,
                 use_hashtag: bool = False,
                 use_senti_tree: bool = False, phrase_num: int = 0, dataset: str = None,
                 data_path: str = None, mode: str = None, num_classes: int = None,
                 train_batch_size: int = 1, eval_batch_size: int = 1, eff_train_batch_size: int = 1, num_workers: int = 0,
                 num_train: int = None, num_dev: int = None, num_test: int = None,
                 num_train_seed: int = None, num_dev_seed: int = None, num_test_seed: int = None,
                 train_shuffle: bool = False

                 ):
        super().__init__()

        self.use_hashtag = use_hashtag
        self.use_senti_tree = use_senti_tree
        self.phrase_num = phrase_num

        self.dataset = dataset
        self.data_path = data_path # ${data_dir}/${.dataset}/${model.arch}/

        self.train_batch_size = train_batch_size
        self.eval_batch_size = eval_batch_size
        self.eff_train_batch_size = eff_train_batch_size
        self.num_workers = num_workers

        self.num_samples = {'train': num_train, 'dev': num_dev, 'test': num_test}
        self.num_samples_seed = {'train': num_train_seed, 'dev': num_dev_seed, 'test': num_test_seed}

        self.train_shuffle = train_shuffle


    def load_dataset(self, split):
        dataset = {}
        data_path = os.path.join(self.data_path, split)
        assert Path(data_path).exists()
        
        for key in tqdm(data_keys, desc=f'Loading {split} set'):
            if self.num_samples[split] is not None:
                filename = f'{key}_{self.num_samples[split]}_{self.num_samples_seed[split]}.pkl'
            else:
                filename = f'{key}.pkl'

            with open(os.path.join(data_path, filename), 'rb') as f:
                dataset[key] = pickle.load(f)

        return dataset

    def setup(self, splits=['all'], stage=None, dataset=None):
        self.data = {}
        if dataset is None:
            splits = ['train', 'dev', 'test'] if splits == ['all'] else splits
            for split in splits:
                dataset = self.load_dataset(split)
                self.data[split] = TextClassificationDataset(dataset, split, self.use_hashtag, self.use_senti_tree, self.phrase_num)
        else:
            self.data['pred'] = TextClassificationDataset(dataset, 'pred', self.use_hashtag, self.use_senti_tree, self.phrase_num)


    def train_dataloader(self):

        return DataLoader(
            self.data['train'],
            batch_size=self.train_batch_size,
            num_workers=self.num_workers,
            collate_fn=self.data['train'].collater,
            shuffle=self.train_shuffle,
            pin_memory=True
        )

    def val_dataloader(self, test=False):
        if test:
            return DataLoader(
                self.data['dev'],
                batch_size=self.eval_batch_size,
                num_workers=self.num_workers,
                collate_fn=self.data['dev'].collater,
                pin_memory=True
            )

        return [
            DataLoader(
            self.data[eval_split],
            batch_size=self.eval_batch_size,
            num_workers=self.num_workers,
            collate_fn=self.data[eval_split].collater,
            pin_memory=True)
            
            for eval_split in ['dev', 'test']
        ]

    def test_dataloader(self):
        return DataLoader(
            self.data['test'],
            batch_size=self.eval_batch_size,
            num_workers=self.num_workers,
            collate_fn=self.data['test'].collater,
            pin_memory=True
        )

    def predict_dataloader(self):
        return DataLoader(
            self.data['pred'],
            batch_size=self.eval_batch_size,
            num_workers=self.num_workers,
            collate_fn=self.data['pred'].collater,
            pin_memory=True
        )

class TextClassificationDataset(Dataset):
    def __init__(self, dataset, split, use_hashtag, use_senti_tree, phrase_num):
        self.data = dataset
        self.split = split
        self.use_hashtag = use_hashtag
        self.use_senti_tree = use_senti_tree
        self.phrase_num = phrase_num

    def __len__(self):
        return len(self.data['item_idx'])

    def __getitem__(self, idx):
        item_idx = torch.LongTensor([self.data['item_idx'][idx]])
        input_ids = torch.LongTensor(self.data['input_ids'][idx])
        attention_mask = torch.LongTensor(self.data['attention_mask'][idx])

        hashtag_ids = torch.LongTensor(self.data['hashtag_ids'][idx])

        result_tuple = (item_idx, input_ids, attention_mask, hashtag_ids)


        if self.use_senti_tree:
            phrase_span_ids = torch.LongTensor(self.data['tree'][idx]['spans'])
            sentiment_ids = torch.LongTensor(self.data['tree'][idx]['sentiment'])
            if phrase_span_ids.shape[0] == 0:
                phrase_span_ids = torch.LongTensor([[0, 0]])
            if phrase_span_ids.shape[1] > self.phrase_num:
                phrase_span_ids = phrase_span_ids[:, :self.phrase_num]
                sentiment_ids = sentiment_ids[:, :self.phrase_num]
            elif phrase_span_ids.shape[1] < self.phrase_num:
                phrase_span_ids = torch.cat([phrase_span_ids, torch.LongTensor([[0, 0]] * (self.phrase_num - phrase_span_ids.shape[0]))])
                sentiment_ids = torch.cat([sentiment_ids, torch.LongTensor([0] * (self.phrase_num - sentiment_ids.shape[0]))])
            result_tuple += (phrase_span_ids, sentiment_ids)

        if 'label' in self.data:
            label = torch.LongTensor([self.data['label'][idx]])
            result_tuple += (label,)
        return result_tuple



    def collater(self, items):

        batch = {
            'item_idx': torch.cat([x[0] for x in items]),
            'input_ids': torch.stack([x[1] for x in items], dim=0),
            'attention_mask': torch.stack([x[2] for x in items], dim=0),
            'hashtag_ids': torch.stack([x[3] for x in items], dim=0) if self.use_hashtag else None,
            'phrase_span_ids': torch.stack([x[4] for x in items]) if self.use_senti_tree else None,
            'sentiment_ids': torch.stack([x[5] for x in items]) if self.use_senti_tree else None,
            'label': torch.cat([x[-1] for x in items]) if self.split != 'pred' else None,
            'split': self.split,
        }
        
        return batch


from transformers import get_scheduler

no_decay = ['bias', 'LayerNorm.weight']


def setup_optimizer_params(model_dict, optimizer, use_hashtag, use_senti_tree, use_emo_cor):
    optimizer_parameters = [
        {
            'params': [p for n, p in model_dict['task_encoder'].named_parameters() if not any(nd in n for nd in no_decay)],
            'weight_decay': optimizer.keywords['weight_decay'],
        },
        {
            'params': [p for n, p in model_dict['task_encoder'].named_parameters() if any(nd in n for nd in no_decay)],
            'weight_decay': 0.0,
        },
        {
            'params': [p for n, p in model_dict['task_head'].named_parameters() if not any(nd in n for nd in no_decay)],
            'weight_decay': optimizer.keywords['weight_decay'],
        },
        {
            'params': [p for n, p in model_dict['task_head'].named_parameters() if any(nd in n for nd in no_decay)],
            'weight_decay': 0.0,
        },
    ]
    if use_hashtag:
        optimizer_parameters += [
            {
                'params': [p for n, p in model_dict['hashtag_encoder'].named_parameters() if not any(nd in n for nd in no_decay)],
                'weight_decay': optimizer.keywords['weight_decay'],
            },
            {
                'params': [p for n, p in model_dict['hashtag_encoder'].named_parameters() if any(nd in n for nd in no_decay)],
                'weight_decay': 0.0,
            }
        ]
    if use_senti_tree:
        optimizer_parameters += [
            {
                'params': [p for n, p in model_dict['phrase_encoder'].named_parameters() if not any(nd in n for nd in no_decay)],
                'weight_decay': optimizer.keywords['weight_decay'],
            },
            {
                'params': [p for n, p in model_dict['phrase_encoder'].named_parameters() if any(nd in n for nd in no_decay)],
                'weight_decay': 0.0,
            },
            {
                'params': [p for n, p in model_dict['phrase_head'].named_parameters() if not any(nd in n for nd in no_decay)],
                'weight_decay': optimizer.keywords['weight_decay'],
            },
            {
                'params': [p for n, p in model_dict['phrase_head'].named_parameters() if any(nd in n for nd in no_decay)],
                'weight_decay': 0.0,
            },
            {
                'params': [p for n, p in model_dict['tok_attn'].named_parameters() if
                           not any(nd in n for nd in no_decay)],
                'weight_decay': optimizer.keywords['weight_decay'],
            },
            {
                'params': [p for n, p in model_dict['tok_attn'].named_parameters() if
                           any(nd in n for nd in no_decay)],
                'weight_decay': 0.0,
            },
            {
                'params': [p for n, p in model_dict['senti_encoder'].named_parameters() if
                           not any(nd in n for nd in no_decay)],
                'weight_decay': optimizer.keywords['weight_decay'],
            },
            {
                'params': [p for n, p in model_dict['senti_encoder'].named_parameters() if
                           any(nd in n for nd in no_decay)],
                'weight_decay': 0.0,
            }
        ]
    if use_emo_cor:
        optimizer_parameters += [
            {
                'params': [p for n, p in model_dict['cor_head'].named_parameters() if
                           not any(nd in n for nd in no_decay)],
                'weight_decay': optimizer.keywords['weight_decay'],
            },
            {
                'params': [p for n, p in model_dict['cor_head'].named_parameters() if
                           any(nd in n for nd in no_decay)],
                'weight_decay': 0.0,
            },
        ]

    return optimizer_parameters

def setup_scheduler(scheduler, total_steps, optimizer):
    if scheduler['warmup_updates'] > 1.0:
        warmup_steps = int(scheduler['warmup_updates'])
    else:
        warmup_steps = int(total_steps *
                            scheduler['warmup_updates'])
    print(
        f'\nTotal steps: {total_steps} with warmup steps: {warmup_steps}\n')

    scheduler = get_scheduler(
        "linear", optimizer=optimizer,
        num_warmup_steps=warmup_steps, num_training_steps=total_steps)

    scheduler = {
        'scheduler': scheduler,
        'interval': 'step',
        'frequency': 1
    }
    return scheduler

def freeze_net(module):
    for p in module.parameters():
        p.requires_grad = False

def unfreeze_net(module):
    for p in module.parameters():
        p.requires_grad = True

def freeze_layers(model, num_freeze_layers):
    if model.arch == 'google/bigbird-roberta-base':
        assert model.task_encoder is not None

        # Freeze task encoder's embedding layer
        for p in model.task_encoder.embeddings.parameters():
            p.requires_grad = False

        # Freeze task encoder's encoder layers
        for i in range(num_freeze_layers):
            for p in model.task_encoder.encoder.layer[i].parameters():
                p.requires_grad = False

        if model.expl_encoder is not None:
            # Freeze expl encoder's embedding layer
            for p in model.expl_encoder.embeddings.parameters():
                p.requires_grad = False

            # Freeze expl encoder's encoder layers
            for i in range(num_freeze_layers):
                for p in model.expl_encoder.encoder.layer[i].parameters():
                    p.requires_grad = False

    else:
        raise NotImplementedError


import torch.nn.functional as F


def calc_task_loss(logits, targets, reduction='mean', class_weights=None, mode='bce'):
    assert len(logits) == len(targets)
    if mode == 'bce':
        return F.binary_cross_entropy_with_logits(logits, targets.float(), weight=class_weights, reduction=reduction)
    elif mode == 'ce':
        return F.cross_entropy(logits, targets, weight=class_weights, reduction=reduction)
    else:
        raise NotImplementedError




#!/usr/bin/env python
import os

from setuptools import find_packages, setup

with open("README.md") as f:
    long_description = f.read()

setup(
    name="multi_emotion",
    version='0.1.13',

    description="detect multiple emotions in a sentence including [anger, anticipation, disgust, fear, joy, love, optimism, pessimism, sadness,\
                    surprise, trust]",
    long_description=long_description,
    long_description_content_type="text/markdown",
    author="Jinfen Li",
    author_email="jli284@syr.edu",
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: Apache Software License",
        'Programming Language :: Python :: 3.9'
    ],
    url="https://github.com/JinfenLi/multi_emotion_recognition",
    license="MIT",
    install_requires=["lightning>=2","torch>=2",
                      "emotlib",
                      "numpy", "pandas", "rich", "torchmetrics>=1",
                      "tqdm",
                      "transformers==4.31.0"],
    packages=find_packages(),


)



"""
    :author: Jinfen Li
    :url: https://github.com/JinfenLi
"""
import json
import os
import pickle
import random
from collections import Counter

import emotlib
import numpy as np
from tqdm import tqdm

from multi_emotion.utils.data import dataset_info, data_keys
import pandas as pd

def nrc_hashtag_lexicon(nrc_lexicon):
    hashtag_vocab = []
    for nl in nrc_lexicon:
        hashtag_vocab.append(nl.split('\t')[1].replace("#", ""))
    return hashtag_vocab


def spanish_hashtag_lexicon(emo_lexicon):
    hashtag_lexicon = []
    for el in emo_lexicon[1:]:
        hashtag_lexicon.append(el.split('\t')[0])
    return hashtag_lexicon

def get_hashtag_inputs(tokens, hashtag_dict):
    hashtag_inputs = []
    cur_word = []
    tid = 0
    while tid < len(tokens):
        if tokens[tid] in ['[PAD]', '[SEP]']:
            hashtag_inputs.extend([0] * (len(tokens) - len(hashtag_inputs)))
            break
        if tid < len(tokens) and tokens[tid].startswith('##'):
            while tid < len(tokens) and tokens[tid].startswith('##'):
                cur_word.append(tokens[tid][2:])
                tid += 1
            # let tid point to the last token of the word
            tid -= 1
        else:
            cur_word = [tokens[tid]]


        if ''.join(cur_word) in hashtag_dict:
            hashtag_id = hashtag_dict[''.join(cur_word)]
            # the hashtags of word: #, xx, ##xx, ##xx are all 1
            if 0 < (tid - len(cur_word)) < len(tokens) and tokens[tid - len(cur_word)] == "#":
                hashtag_inputs[-1] = hashtag_id
            hashtag_inputs.extend([hashtag_id] * len(cur_word))
        else:
            hashtag_inputs.extend([0] * len(cur_word))

        tid += 1
    return hashtag_inputs


def update_dataset_dict(
    idx, dataset_dict, input_ids, hashtag_inputs, max_length, tokenizer, text, token_offsets, label=None):
    input_ids = [tokenizer.cls_token_id] + input_ids + [tokenizer.sep_token_id]
    hashtag_inputs = [0] + hashtag_inputs + [0]
    num_tokens = len(input_ids)
    truncated_pos = token_offsets[-1][1] if text else 0

    num_pad_tokens = max_length - num_tokens
    assert num_pad_tokens >= 0

    input_ids += [tokenizer.pad_token_id] * num_pad_tokens
    attention_mask = [1] * num_tokens + [0] * num_pad_tokens
    hashtag_inputs += [0] * num_pad_tokens
    if token_offsets is not None:
        token_offsets = [(0, 0)] + token_offsets + [(0, 0)]
        token_offsets += [(0, 0)] * num_pad_tokens


    dataset_dict['item_idx'].append(idx)
    dataset_dict['input_ids'].append(input_ids)
    dataset_dict['hashtag_ids'].append(hashtag_inputs)
    dataset_dict['attention_mask'].append(attention_mask)
    if label is not None:
        dataset_dict['label'].append(label)
    dataset_dict['offsets'].append(token_offsets)
    dataset_dict['truncated_texts'].append(text[: truncated_pos])

    return dataset_dict

def preprocess_dataset(text):
    text = text.replace("\\n", " ")
    text = text.lower()
    text = emotlib.demojify(text)
    return text

def transform_data(tokenizer, hashtag_dict, text, max_length):
    encode_dict = tokenizer(text, add_special_tokens=False, return_offsets_mapping=True)
    raw_tokens = encode_dict.encodings[0].tokens
    hashtag_inputs = get_hashtag_inputs(raw_tokens, hashtag_dict)
    # num_tokens = encode_dict['attention_mask'].count(1)
    input_ids = encode_dict['input_ids']
    offset_mapping = encode_dict['offset_mapping']
    # if num_tokens > actual_max_length:
    #     actual_max_length = num_tokens

    input_length = min(len(input_ids), max_length - 2)
    input_ids = input_ids[:input_length]
    offset_mapping = offset_mapping[:input_length]
    hashtag_inputs = hashtag_inputs[:input_length]
    return input_ids, hashtag_inputs, offset_mapping

def generate_output_file(dataset, tokenizer, input_ids, probabilities, output_dir=None, targets=None):
    texts = []
    for i in range(len(input_ids)):
        text = tokenizer.decode(input_ids[i], skip_special_tokens=True)
        texts.append(text)
    labels = [','.join([dataset_info[dataset]['classes'][i] for i, pp in enumerate(p) if pp > 0.5]) for p in probabilities]
    labels = ['neutral' if len(l) == 0 else l for l in labels]
    probs = [json.dumps([{dataset_info[dataset]['classes'][i]: pp.item()} for i, pp in enumerate(p)]) for p in probabilities]
    results = [{'text': t, 'pred_label': l, 'probability': p} for t, l, p in zip(texts, labels, probs)]
    result_df = pd.DataFrame({'text': texts, 'prediction': labels, 'prob': probs})
    if targets is not None:
        targets = [','.join([dataset_info[dataset]['classes'][i] for i, tt in enumerate(t) if tt != 0]) for t in
                   targets]
        result_df['target'] = targets
    if output_dir:
        result_df.to_csv(os.path.join(output_dir, 'result.csv'), index=False)
    return results

def stratified_sampling(data, num_samples):
    num_instances = len(data)
    assert num_samples < num_instances

    counter_dict = Counter(data)
    unique_vals = list(counter_dict.keys())
    val_counts = list(counter_dict.values())
    num_unique_vals = len(unique_vals)
    assert num_unique_vals > 1

    num_stratified_samples = [int(c*num_samples/num_instances) for c in val_counts]
    assert sum(num_stratified_samples) <= num_samples
    if sum(num_stratified_samples) < num_samples:
        delta = num_samples - sum(num_stratified_samples)
        delta_samples = np.random.choice(range(num_unique_vals), replace=True, size=delta)
        for val in delta_samples:
            num_stratified_samples[unique_vals.index(val)] += 1
    assert sum(num_stratified_samples) == num_samples

    sampled_indices = []
    for i, val in enumerate(unique_vals):
        candidates = np.where(data == val)[0]
        sampled_indices += list(np.random.choice(candidates, replace=False, size=num_stratified_samples[i]))
    random.shuffle(sampled_indices)

    return sampled_indices
def sample_dataset(data_path, dataset_dict, split, num_samples, seed):
    sampled_split_filename = f'{split}_split_{num_samples}_{seed}.pkl'
    if os.path.exists(os.path.join(data_path, sampled_split_filename)):
        with open(os.path.join(data_path, sampled_split_filename), 'rb') as f:
            sampled_split = pickle.load(f)
    else:
        sampled_split = stratified_sampling(dataset_dict['label'], num_samples)
        with open(os.path.join(data_path, sampled_split_filename), 'wb') as f:
            pickle.dump(sampled_split, f)

    for key in data_keys:
        dataset_dict[key] = sampled_split if key == 'item_idx' else [dataset_dict[key][i] for i in sampled_split]

    return dataset_dict

def save_datadict(data_path, dataset_dict, split, num_samples, seed):
    for key in tqdm(data_keys, desc=f'Saving {split} dataset'):
        if key in dataset_dict:
            filename = f'{key}.pkl' if num_samples is None else f'{key}_{num_samples}_{seed}.pkl'
            with open(os.path.join(data_path, filename), 'wb') as f:
                pickle.dump(dataset_dict[key], f)


import torch
from torchmetrics.classification import BinaryAccuracy, BinaryF1Score
def init_best_metrics():
    return {
        'best_epoch': 0,
        'dev_best_perf': None,
        'test_best_perf': None,
    }

def init_perf_metrics(num_classes):
    # if num_classes > 2:
    #     task: Literal["multiclass"] = "multiclass"
    # else:
    #     task: Literal["binary"] = "binary"
    perf_metrics = torch.nn.ModuleDict({
        'acc': BinaryAccuracy(),
        'macro_f1': BinaryF1Score(),
        # 'micro_f1': BinaryF1Score(multidim_average='samplewise'),
    })

    # assert num_classes >= 2
    return perf_metrics

def calc_preds(logits):
    probabilities = logits.sigmoid()
    # binary_labels = (probabilities > .5).int()
    return probabilities


def get_step_metrics(preds, targets, metrics):
    res = {}
    for key, metric_fn in metrics.items():
        res.update({key: metric_fn(preds, targets) * 100})
    return res

def get_epoch_metrics(metrics):
    res = {}
    for key, metric_fn in metrics.items():
        res.update({key: metric_fn.compute() * 100})
        metric_fn.reset()
    return res